# model.py
import json
import numpy as np
import triton_python_backend_utils as pb_utils
import cv2
import random
import math
from enum import Enum
from torch.utils.dlpack import from_dlpack
import time
import uuid
import os
import torch
import threading
import base64


with open("/images/imagenet_labels/imagenet_classes_1328.txt") as f:
    classes = [line.strip() for line in f.readlines()]

with open("/images/imagenet_labels/names_cls1328.txt") as f:
    names = eval(f.readlines()[0].strip())

configs_dic = {}
with open("/images/configs/model_configs.txt") as f:
    data_list = f.readlines()
    for data in data_list:
        if len(data)==0:
            continue
        data = data.strip()
        key = data.split(":")[0].strip()
        value = data.split(":")[1].strip()
        if value[0].isnumeric():   # configs
            configs_dic[key] = float(value)
        else:
            configs_dic[key] = value

if bool(configs_dic["plot_bool"]):
    _LINE_THICKNESS_SCALING = 500.0
    _TEXT_THICKNESS_SCALING = 700.0
    _TEXT_SCALING = 520.0
    np.random.seed(69)
    RAND_COLORS = np.random.randint(10, 255, (int(configs_dic['cls_num']), 3), "int")  # used for class visu  # 1328 classes nun color


class BoundingBox:
    def __init__(self, classID, confidence, x1, x2, y1, y2, image_height, image_width): # erro is raw image_height, image_width
        self.classID = int(classID)
        self.confidence = float(confidence)
        self.x1 = x1 = float(x1)
        self.x2 = x2 = float(x2)
        self.y1 = y1 = float(y1)
        self.y2 = y2 = float(y2)
        image_width = float(image_width)
        image_height = float(image_height)
        self.u1 = x1 / image_width
        self.u2 = x2 / image_width
        self.v1 = y1 / image_height
        self.v2 = y2 / image_height

    def box(self):
        return (self.x1, self.y1, self.x2, self.y2)

    def width(self):
        return self.x2 - self.x1

    def height(self):
        return self.y2 - self.y1

    def center_absolute(self):
        return (0.5 * (self.x1 + self.x2), 0.5 * (self.y1 + self.y2))

    def center_normalized(self):
        return (0.5 * (self.u1 + self.u2), 0.5 * (self.v1 + self.v2))

    def size_absolute(self):
        return (self.x2 - self.x1, self.y2 - self.y1)

    def size_normalized(self):
        return (self.u2 - self.u1, self.v2 - self.v1)


class inferThread(threading.Thread):
    def __init__(self,triton_self,text):
        threading.Thread.__init__(self)
        self.triton_self = triton_self
        self.text = text

    def run(self):
        self.detect_cls_str = self.triton_self.infer(self.text)

    def get_result(self):
        threading.Thread.join(self)
        try:
            return self.detect_cls_str
        except Exception:
            return json.dumps([])

class TritonPythonModel:
    def initialize(self, args):
        self.model_config = model_config = json.loads(args['model_config'])
        output0_config = pb_utils.get_output_config_by_name(model_config, "output0")
        self.output0_dtype = pb_utils.triton_string_to_numpy(output0_config['data_type'])

    def execute(self, requests):
        output0_dtype = self.output0_dtype
        responses = []
        for request in requests:
            in_0 = pb_utils.get_input_tensor_by_name(request, 'input0')
            in_0 = in_0.as_numpy()   #获取输入字符串的numpy格式数据
            text = in_0[0].decode("utf-8")   #输入的第0个字符，即传入的数据，该数据是客户端编码的格式，需要解码。 '/images/input_img/4.jpg'
            thread1 = inferThread(self,text)
            thread1.start()
            thread1.join()
            detect_cls_str = thread1.get_result()
            out_str_num_name = np.array(detect_cls_str)
            out_tensor_0 = pb_utils.Tensor('output0', out_str_num_name.astype(output0_dtype))
            inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_0])
            responses.append(inference_response)
        return responses

    def finalize(self):
        print('Cleaning up...')

    def infer(self,text):
        detected_objects_, img_ = self._recognize_detect(text)
        detect_cls_list = []
        count_obj_num = 1
        if len(detected_objects_) != 0:
            for box in detected_objects_:
                x1, y1, x2, y2 = box.box()
                w = box.width()
                h = box.height()
                detect_classID = box.classID
                resize_scale = 640 / max(img_.shape[0],img_.shape[1])
                limite_letter_box = configs_dic['limite_letter_box']
                if (w * resize_scale) >= limite_letter_box and (h * resize_scale) >= limite_letter_box and detect_classID == 0:  # 限制最小识别目标
                    if w != 224 and h != 224:
                        resize_cls_image = self._resize_cls(x1, y1, w, h, img_,)
                    else:
                        resize_cls_image = img_
                    classes_ls_, predicted_conf_ls_, category_id_ls_ = self._recognize_cls(resize_cls_image)
                    if predicted_conf_ls_[0] >= configs_dic['cls_conf']:  # 限制cls的置信度
                        tmp_dic = {}
                        tmp_dic['object'] = count_obj_num
                        tmp_dic['box'] = box.box()
                        tmp_dic['classID'] = category_id_ls_[0]
                        tmp_dic['bird_name'] = classes_ls_
                        tmp_dic['bird_conf'] = predicted_conf_ls_
                        detect_cls_list.append(tmp_dic)
                        count_obj_num += 1
        # save json step 1
        detect_cls_json = json.dumps(detect_cls_list)

        # plot images module
        if bool(configs_dic["plot_bool"]):
            if len(detect_cls_list) != 0:
                for data in detect_cls_list:
                    # plot
                    img_ = self.render_box(img_, data['box'], color=tuple(RAND_COLORS[data['classID'] % int(configs_dic['cls_num'])].tolist()))    #  plot classes num color
                    size = self.get_text_size(img_, "{}:{}:{:.2f}".format(data['object'],data['bird_name'][0],data['bird_conf'][0]),normalised_scaling=0.6)
                    img_ = self.render_filled_box(img_,(data['box'][0] - 3, data['box'][1] + 3, data['box'][0] + size[0], data['box'][1] - size[1]),color=tuple(RAND_COLORS[data['classID'] % int(configs_dic['cls_num'])].tolist()))
                    img_ = self.render_text(img_, "{}:{}:{:.2f}".format(data['object'],data['bird_name'][0],data['bird_conf'][0]),
                                              (data['box'][0], data['box'][1] - size[1]), color=(225, 255, 255),normalised_scaling=0.6)

            # 文件夹管理2（根据日期建立文件夹，根据时间输出图片和json文件）
            time_now = time.strftime("%Y%m%d", time.localtime())
            output_image_path = '/images/output_img/' + time_now
            if not os.path.exists(output_image_path):
                os.makedirs(output_image_path)
            # save json step 2
            # output_json_path = '/jsons/' + time_now
            # if not os.path.exists(output_json_path):
            #     os.makedirs(output_json_path)

            uuidFour = str(uuid.uuid4())
            out_pic_name = output_image_path + '/' + uuidFour + '.jpg'
            if img_ is not None:
                cv2.imwrite(out_pic_name, img_.astype(np.uint8))
            # save json step 3
            # out_json_name = output_json_path + '/' + uuidFour + '.json'
            # out_pic_json_name = out_pic_name + " " + out_json_name

            #  save json step 4
            # with open(out_json_name,"w") as f:
            #     f.write(detect_cls_json)
        # return out_pic_json_name,detect_cls_json
        return detect_cls_json

    def _recognize_detect(self,draw_path):
        con_thres = configs_dic['det_conf']  # 置信度阈值
        iou_thres = configs_dic['det_iou']  # 非极大值抑制阈值 iou
        img_size = [640, 640]
        text = draw_path
        if text[:4] == "/9j/":
            # type 1 is base64
            # base64 base64_str to base64 to opencv numpy
            base64_data = text.encode('utf8')
            img_base64 = base64.b64decode(base64_data)
            np_data = np.frombuffer(img_base64, np.uint8)
            img_arr = cv2.imdecode(np_data, cv2.IMREAD_COLOR)
            draw = img_arr
        else:
            # type 2 is base64
            draw = cv2.imread(text)
        
        if draw is None or draw.shape[-1] == 4:
            return [],None
        
        if draw.shape[0] < 40 and draw.shape[1] < 40:
            return [],None

        # only cls,not detect
        if draw.shape[0] == 224 and draw.shape[1] == 224:
            detected_objects = []
            label,score,origin_h,origin_w = np.int64(0),np.float32(1),int(224),int(224)
            x1, x2, y1, y2 = np.float32(0),np.float32(224),np.float32(0),np.float32(224)
            detected_objects.append(BoundingBox(label, score, x1, x2, y1, y2, origin_h, origin_w))
            return detected_objects, draw

        input_image_buffer = self.preprocess_detect(draw, img_size)
        inference_request = pb_utils.InferenceRequest(
            model_name=configs_dic['detect_model'],
            requested_output_names=['prob'],
            inputs=[pb_utils.Tensor('data', input_image_buffer)]
        )
        inference_response = inference_request.exec()
        result = self.pb_tensor_to_numpy(pb_utils.get_output_tensor_by_name(inference_response, 'prob'))
        detected_objects = self.postprocess_detect(result, draw.shape[1], draw.shape[0], img_size, con_thres, iou_thres)
        return detected_objects,draw

    def _recognize_cls(self, resize_cls_image):
        input_image_buffer = self.preprocess_cls(resize_cls_image)
        inference_request = pb_utils.InferenceRequest(
            model_name=configs_dic['cls_model'],
            requested_output_names=['prob'],
            inputs=[pb_utils.Tensor('data', input_image_buffer)]
        )
        inference_response = inference_request.exec()
        result = self.pb_tensor_to_numpy(pb_utils.get_output_tensor_by_name(inference_response, 'prob'))
        result = result.reshape(int(configs_dic['cls_num']))  # classes num
        classes_ls, predicted_conf_ls, category_id_ls = self.postprocess_cls(result,top_k=5)  # [,,,,,],[,,,,,],[,,,,,]
        return classes_ls, predicted_conf_ls, category_id_ls

    def _resize_cls(self, x1, y1, w, h,im0, new_shape=(224, 224), color=(114, 114, 114)):
        # crop image
        x1, y1, w, h = int(x1), int(y1), int(w), int(h)
        im = im0[y1:y1 + h, x1:x1 + w, :]
        # Resize and pad image while meeting stride-multiple constraints
        shape = im.shape[:2]  # current shape [height, width]
        # Scale ratio (new / old)
        r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])

        # Compute padding
        new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
        dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
        dw /= 2  # divide padding into 2 sides
        dh /= 2
        if shape[::-1] != new_unpad:  # resize
            im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
        top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
        left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
        im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
        return im

    def pb_tensor_to_numpy(self,pb_tensor):
        '''pb_tensor转换为numpy格式'''
        if pb_tensor.is_cpu():
            return pb_tensor.as_numpy()
        else:
            pytorch_tensor = from_dlpack(pb_tensor.to_dlpack())
            return pytorch_tensor.cpu().numpy()

    def preprocess_detect(self, raw_bgr_image, input_shape):
        input_w, input_h = input_shape
        image_raw = raw_bgr_image
        h, w, c = image_raw.shape
        image = cv2.cvtColor(image_raw, cv2.COLOR_BGR2RGB)
        # Calculate widht and height and paddings
        r_w = input_w / w
        r_h = input_h / h
        if r_h > r_w:
            tw = input_w
            th = int(r_w * h)
            tx1 = tx2 = 0
            ty1 = int((input_h - th) / 2)
            ty2 = input_h - th - ty1
        else:
            tw = int(r_h * w)
            th = input_h
            tx1 = int((input_w - tw) / 2)
            tx2 = input_w - tw - tx1
            ty1 = ty2 = 0
        # Resize the image with long side while maintaining ratio
        image = cv2.resize(image, (tw, th))
        # Pad the short side with (128,128,128)
        image = cv2.copyMakeBorder(
            image, ty1, ty2, tx1, tx2, cv2.BORDER_CONSTANT,None, (128, 128, 128)
        )
        image = image.astype(np.float32)
        # Normalize to [0,1]
        image /= 255.0
        # HWC to CHW format:
        image = np.transpose(image, [2, 0, 1])
        image = np.expand_dims(image, axis=0)
        image = np.ascontiguousarray(image)
        return image

    def preprocess_cls(self,input_img):
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
        im = cv2.cvtColor(input_img, cv2.COLOR_BGR2RGB)
        im = np.float32(im)
        im /= 255.0
        im -= mean
        im /= std
        im = im.transpose(2, 0, 1)
        # prepare batch
        batch_data = np.expand_dims(im, axis=0)
        batch_data = np.ascontiguousarray(batch_data)
        return batch_data

    def postprocess_detect(self, output, origin_w, origin_h, input_shape, conf_th=0.5, nms_threshold=0.4):
        """Postprocess TensorRT outputs.
        # Args
            output: list of detections with schema
            [num_boxes,cx,cy,w,h,conf,cls_id, cx,cy,w,h,conf,cls_id, ...]
            conf_th: confidence threshold
        # Returns
            list of bounding boxes with all detections above threshold and after nms, see class BoundingBox
        """
        # Get the num of boxes detected
        # Here we use the first row of output in that batch_size = 1
        output = output[0]
        num = int(output[0])
        if num == 0:
            return []
        # Reshape to a two dimentional ndarray
        pred = np.reshape(output[1:], (-1, 6))[:num, :]

        # Do nms
        boxes = self.non_max_suppression(pred, origin_h, origin_w, input_shape[0], input_shape[1], conf_thres=conf_th, nms_thres=nms_threshold)
        if boxes.shape[0] == 0:
            return []
        result_boxes = boxes[:, :4]
        result_scores = boxes[:, 4]
        result_classid = boxes[:, 5].astype(np.int)

        detected_objects = []
        for box, score, label in zip(result_boxes, result_scores, result_classid):
            detected_objects.append(BoundingBox(label, score, box[0], box[2], box[1], box[3], origin_h, origin_w))
        return detected_objects

    def postprocess_cls(self,output_data,top_k):
        classes_ls = []
        predicted_conf_ls = []
        category_id_ls = []
        output_data = output_data.reshape(1, -1)
        output_data = torch.Tensor(output_data)
        p = torch.nn.functional.softmax(output_data, dim=1)
        score, index = torch.topk(p, top_k)
        # myself add
        remake_list = []
        for re_index in index[0]:
            remake_list.append(int(names[int(re_index)]))
        index = torch.unsqueeze(torch.tensor(remake_list), dim=0)
        for ind in range(index.shape[0]):
            for top_k_i in range(top_k):
                input_category_id = index[ind][top_k_i].item()  # 716
                category_id_ls.append(input_category_id)
                predicted_confidence = score[ind][top_k_i].item()
                predicted_conf_ls.append(predicted_confidence)
                classes_ls.append(classes[input_category_id])
        return classes_ls, predicted_conf_ls, category_id_ls

    def non_max_suppression(self, prediction, origin_h, origin_w, input_w, input_h, conf_thres=0.5, nms_thres=0.4):
        """
        description: Removes detections with lower object confidence score than 'conf_thres' and performs
        Non-Maximum Suppression to further filter detections.
        param:
            prediction: detections, (x1, y1, x2, y2, conf, cls_id)
            origin_h: original image height
            origin_w: original image width
            conf_thres: a confidence threshold to filter detections
            nms_thres: a iou threshold to filter detections
        return:
            boxes: output after nms with the shape (x1, y1, x2, y2, conf, cls_id)
        """
        # Get the boxes that score > CONF_THRESH
        boxes = prediction[prediction[:, 4] >= conf_thres]
        # Trandform bbox from [center_x, center_y, w, h] to [x1, y1, x2, y2]
        boxes[:, :4] = self.xywh2xyxy(boxes[:, :4], origin_h, origin_w, input_w, input_h)
        # clip the coordinates
        boxes[:, 0] = np.clip(boxes[:, 0], 0, origin_w - 1)
        boxes[:, 2] = np.clip(boxes[:, 2], 0, origin_w - 1)
        boxes[:, 1] = np.clip(boxes[:, 1], 0, origin_h - 1)
        boxes[:, 3] = np.clip(boxes[:, 3], 0, origin_h - 1)
        # Object confidence
        confs = boxes[:, 4]
        # Sort by the confs
        boxes = boxes[np.argsort(-confs)]
        # Perform non-maximum suppression
        keep_boxes = []
        while boxes.shape[0]:
            large_overlap = self.bbox_iou(np.expand_dims(boxes[0, :4], 0), boxes[:, :4]) > nms_thres
            label_match = boxes[0, -1] == boxes[:, -1]
            # Indices of boxes with lower confidence scores, large IOUs and matching labels
            invalid = large_overlap & label_match
            keep_boxes += [boxes[0]]
            boxes = boxes[~invalid]
        boxes = np.stack(keep_boxes, 0) if len(keep_boxes) else np.array([])
        return boxes

    def xywh2xyxy(self, x, origin_h, origin_w, input_w, input_h):
        """
        description:    Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
        param:
            origin_h:   height of original image
            origin_w:   width of original image
            x:          A boxes numpy, each row is a box [center_x, center_y, w, h]
        return:
            y:          A boxes numpy, each row is a box [x1, y1, x2, y2]
        """
        y = np.zeros_like(x)
        r_w = input_w / origin_w
        r_h = input_h / origin_h
        if r_h > r_w:
            y[:, 0] = x[:, 0] - x[:, 2] / 2
            y[:, 2] = x[:, 0] + x[:, 2] / 2
            y[:, 1] = x[:, 1] - x[:, 3] / 2 - (input_h - r_w * origin_h) / 2
            y[:, 3] = x[:, 1] + x[:, 3] / 2 - (input_h - r_w * origin_h) / 2
            y /= r_w
        else:
            y[:, 0] = x[:, 0] - x[:, 2] / 2 - (input_w - r_h * origin_w) / 2
            y[:, 2] = x[:, 0] + x[:, 2] / 2 - (input_w - r_h * origin_w) / 2
            y[:, 1] = x[:, 1] - x[:, 3] / 2
            y[:, 3] = x[:, 1] + x[:, 3] / 2
            y /= r_h

        return y

    def bbox_iou(self, box1, box2, x1y1x2y2=True):
        """
        description: compute the IoU of two bounding boxes
        param:
            box1: A box coordinate (can be (x1, y1, x2, y2) or (x, y, w, h))
            box2: A box coordinate (can be (x1, y1, x2, y2) or (x, y, w, h))
            x1y1x2y2: select the coordinate format
        return:
            iou: computed iou
        """
        if not x1y1x2y2:
            # Transform from center and width to exact coordinates
            b1_x1, b1_x2 = box1[:, 0] - box1[:, 2] / 2, box1[:, 0] + box1[:, 2] / 2
            b1_y1, b1_y2 = box1[:, 1] - box1[:, 3] / 2, box1[:, 1] + box1[:, 3] / 2
            b2_x1, b2_x2 = box2[:, 0] - box2[:, 2] / 2, box2[:, 0] + box2[:, 2] / 2
            b2_y1, b2_y2 = box2[:, 1] - box2[:, 3] / 2, box2[:, 1] + box2[:, 3] / 2
        else:
            # Get the coordinates of bounding boxes
            b1_x1, b1_y1, b1_x2, b1_y2 = box1[:, 0], box1[:, 1], box1[:, 2], box1[:, 3]
            b2_x1, b2_y1, b2_x2, b2_y2 = box2[:, 0], box2[:, 1], box2[:, 2], box2[:, 3]

        # Get the coordinates of the intersection rectangle
        inter_rect_x1 = np.maximum(b1_x1, b2_x1)
        inter_rect_y1 = np.maximum(b1_y1, b2_y1)
        inter_rect_x2 = np.minimum(b1_x2, b2_x2)
        inter_rect_y2 = np.minimum(b1_y2, b2_y2)
        # Intersection area
        inter_area = np.clip(inter_rect_x2 - inter_rect_x1 + 1, 0, None) * \
                    np.clip(inter_rect_y2 - inter_rect_y1 + 1, 0, None)
        # Union Area
        b1_area = (b1_x2 - b1_x1 + 1) * (b1_y2 - b1_y1 + 1)
        b2_area = (b2_x2 - b2_x1 + 1) * (b2_y2 - b2_y1 + 1)

        iou = inter_area / (b1_area + b2_area - inter_area + 1e-16)

        return iou

    def render_box(self, img, box, color=(200, 200, 200)):
        """
        Render a box. Calculates scaling and thickness automatically.
        :param img: image to render into
        :param box: (x1, y1, x2, y2) - box coordinates
        :param color: (b, g, r) - box color
        :return: updated image
        """
        x1, y1, x2, y2 = box
        thickness = int(
            round(
                (img.shape[0] * img.shape[1])
                / (_LINE_THICKNESS_SCALING * _LINE_THICKNESS_SCALING)
            )
        )
        thickness = max(1, thickness)
        img = cv2.rectangle(
            img,
            (int(x1), int(y1)),
            (int(x2), int(y2)),
            color,
            thickness=thickness
        )
        return img

    def render_filled_box(self, img, box, color=(200, 200, 200)):
        """
        Render a box. Calculates scaling and thickness automatically.
        :param img: image to render into
        :param box: (x1, y1, x2, y2) - box coordinates
        :param color: (b, g, r) - box color
        :return: updated image
        """
        x1, y1, x2, y2 = box
        img = cv2.rectangle(
            img,
            (int(x1), int(y1)),
            (int(x2), int(y2)),
            color,
            thickness=cv2.FILLED
        )
        return img

    def get_text_size(self, img, text, normalised_scaling=1.0):
        """
        Get calculated text size (as box width and height)
        :param img: image reference, used to determine appropriate text scaling
        :param text: text to display
        :param normalised_scaling: additional normalised scaling. Default 1.0.
        :return: (width, height) - width and height of text box
        """
        thickness = int(
            round(
                (img.shape[0] * img.shape[1])
                / (_TEXT_THICKNESS_SCALING * _TEXT_THICKNESS_SCALING)
            )
            * normalised_scaling
        )
        thickness = max(1, thickness)
        scaling = img.shape[0] / _TEXT_SCALING * normalised_scaling
        return cv2.getTextSize(text, cv2.FONT_HERSHEY_SIMPLEX, scaling, thickness)[0]


    def render_text(self, img, text, pos, color=(200, 200, 200), normalised_scaling=1.0):
        """
        Render a text into the image. Calculates scaling and thickness automatically.
        :param img: image to render into
        :param text: text to display
        :param pos: (x, y) - upper left coordinates of render position
        :param color: (b, g, r) - text color
        :param normalised_scaling: additional normalised scaling. Default 1.0.
        :return: updated image
        """
        x, y = pos
        thickness = int(
            round(
                (img.shape[0] * img.shape[1])
                / (_TEXT_THICKNESS_SCALING * _TEXT_THICKNESS_SCALING)
            )
            * normalised_scaling
        )
        thickness = max(1, thickness)
        scaling = img.shape[0] / _TEXT_SCALING * normalised_scaling
        size = self.get_text_size(img, text, normalised_scaling)
        cv2.putText(
            img,
            text,
            (int(x), int(y + size[1])),
            cv2.FONT_HERSHEY_SIMPLEX,
            scaling,
            color,
            thickness=thickness,
            lineType=cv2.LINE_AA,
        )
        return img
